# Resiti sledeci optimizacioni problem:
# min  x*y^2 - x^2 + y^2
# s.t. x + y >= 4
#      x*y <= 5
#      1 <= x <= 3
#      2 <= y <= 3

# Jasno je da se radi o problemu nelinearnog progrmairanja
# (nelinearna funkcija cilja i nelinearna ogranicenja),
# pa moramo da koristimo minimize metod iz scipy.optimize
# paketa
import numpy as np
import scipy.optimize as scp
import matplotlib.pyplot as plt

# kodiramo funkciju cilja
def objective(x):
    x2 = x[0]*x[0]
    y2 = x[1]*x[1]
    return x[0] * y2 - x2 - y2

# kodiramo ogranicenje: x + y >= 4
# prebacujemo ga u oblik f(x) = x + y - 4 >= 0
def constraint1(x):
    return x[0] + x[1] - 4

# kodiramo ogranicenje: x*y <= 5
# prebacujemo ga u oblik f(x) = 5 - x*y >= 0
def constraint2(x):
    return 5 - x[0]*x[1]

# kodiramo intervale za promenljive
bx = (1,3)
by = (2,3)
bnds = (bx, by)
# kodiramo ogranicenja
cons = [
    {"type" : "ineq", "fun" : constraint1},
    {"type" : "ineq", "fun" : constraint2},
]
# definisemo dopustivo pocetno resenje
x0 = [1,2]
# definisemo maks broj iteracija
opt = {"maxiter" : 1000}

# resavamo zadatak
res = scp.minimize(objective, x0, constraints = cons, bounds = bnds, method = "SLSQP", options = opt)

# prikazujemo resenje
print(f"Optimalno resenje: x = {res.x[0]}, y = {res.x[1]}")
print(f"Maksimalna vrednost funkcije: {res.fun}")

# mozemo da vizuelizujemo resenje
x_opt = res.x[0]
y_opt = res.x[1]

# kreiramo podele
xx = np.linspace(0,4, num=200)
yy = np.linspace(0,4, num=200)
x,y = np.meshgrid(xx, yy)
obj = x*y*y -x*x - y*y

# crtamo figuru
plt.figure(1)
# crtamo skup dopustivih resenja
# radi lepse slike su ignorasana stvarna ogranicenja za x i y
plt.imshow(((x+y >=4) & (x*y <= 5) & (x >= 0) & (y >= 0)).astype(int),
           extent = (x.min(), x.max(), y.min(), y.max()), origin="lower", cmap="Greys", alpha=0.3)
# osencicemo stvarni dopustivi skup
plt.imshow(((x+y >=4) & (x*y <= 5) & (x >= 1) & (x <= 3) & (y >= 2) & (y <= 3)).astype(int),
           extent = (x.min(), x.max(), y.min(), y.max()), origin="lower", cmap="Reds", alpha=0.3)

# crtamo ogranicenja
x0 = np.linspace(1,3,num=200)
y1 = 4 - x0 # x + y >= 4
y2 = 5.0/x0 # x*y <= 5
y3 = np.ones(len(x0))*3 # y <= 3
y4 = np.ones(len(x0))*2 # y >= 1
plt.plot(x0, y1, 'r--', linewidth=2,label=r'$x+y\geq4$')
plt.plot(x0, y2, 'k--', linewidth=2, label=r"$xy \leq 5$")
plt.plot(x0, y3, 'b--', linewidth=2, label=r"$y \leq 3$")
plt.plot(x0, y4, 'g--', linewidth=2, label=r"$y \geq 2$")
plt.plot([1,1],[2,3],'o--',label=r'$x\geq1$')
plt.plot([3,3],[2,3],'m--',label=r'$x\leq3$')

# optimalno resenje
plt.plot([x_opt], [y_opt], marker='o', color='orange', markersize=10)


# dodajemo konture funkcije cilja
# konture funkcije cilja
cs = plt.contour(x,y,obj, np.arange(-5,5))
plt.clabel(cs, inline=1, fontsize=20)

plt.title("Nelinearno programiranje")
plt.xlabel("x")
plt.ylabel("y")
plt.legend(loc="upper right")
plt.grid()

plt.tight_layout()
plt.show()